
open Stree

module Symtab = Map.Make (StreeOrder)

exception Bad_args of string
exception Pattern_match
let null = Symbol "nil"

let rec builtin_print n = function
        | Seq s -> if n > 0 then print_string "("; List.iter (fun elem ->
        builtin_print (n+1) elem; print_string " ") s; if n > 0 then
        print_string ")" else print_newline ()
        | Symbol s -> print_string s
        | x -> print_string (List.assoc x op_names)

let rec print_symtab symtab =
        Symtab.iter (fun key elem -> builtin_print 0 key;
        print_string " ---> "; builtin_print 0 elem) symtab

let (>>) f g x = f (g x)
let id x = x
let fn_cons i = fun t -> i :: t
let fn_append l = fun t -> l @ t
let fn_to_normal f = f []

let rec match_unify patt value symtab =
        match (patt, value) with
        | Seq [Symbol _ as name], Seq _ -> Symtab.add name value symtab
        | Seq (p::ps), Seq (d::ds) ->
            match_unify (Seq ps) (Seq ds) (match_unify p d symtab)
        | Seq [], Seq [] -> symtab
        | Symbol _, _ -> Symtab.add patt value symtab
        | _, _ -> raise Pattern_match

let fill_template templ symtab =
    let rec fill = function
        | Symbol _ as s ->
        begin
            try Symtab.find s symtab
            with Not_found -> s
        end
        | Seq s -> Seq (List.map fill s)
        | x -> x
    in fill templ

let rec execute_i k acc symtab seq =
        let continue res rest = execute_i k (acc >> res) symtab rest in
        match seq with

        | Symbol n as s :: rest ->
            let value = try Symtab.find s symtab
            with Not_found -> null
            in continue (fn_cons value) rest

        | Seq s :: rest -> continue (fn_append s) rest

        | Print :: arg :: rest -> builtin_print 0 arg; continue id rest

        | Input :: rest ->
            continue (fn_append (Syntax.from_string (read_line ()))) rest

        | Quote :: arg :: rest -> continue (fn_cons (Seq [arg])) rest

        | Define :: patt :: data :: rest ->
            let new_symtab = match_unify patt data symtab in
            execute_i k acc new_symtab rest

        | Concat :: arg1 :: arg2 :: rest ->
            let items1 = match arg1 with Seq x -> x | x -> [x]
            and items2 = match arg2 with Seq x -> x | x -> [x] in
            continue (fn_cons (Seq (items1 @ items2))) rest

        | Ipperopp :: rest ->
            let old_symtab acc _ = k acc symtab in
            execute_i old_symtab acc symtab rest

        | Apply :: value :: patt :: result :: rest ->
            let tmp_symtab = match_unify patt value Symtab.empty in
            if !Options.debug then print_symtab tmp_symtab;
            continue (fn_cons (fill_template result tmp_symtab)) rest

        | Execute :: Seq arg :: rest ->
            (* first evaluate the requested seq.  Then re-evaluate and
               collect the results together in the containing sequence. *)
            let stage2 acc2 symtab =
                let rest = acc2 rest in
                if !Options.debug then begin
                        print_string "Re-eval: ";
                        builtin_print 0 (Seq rest)
                end;
                execute_i k acc symtab rest in
            if !Options.debug then begin
                    print_string "Step in: ";
                    builtin_print 0 (Seq arg)
            end;
            execute_i stage2 id symtab arg

        | Debug :: rest ->
                print_symtab symtab;
                continue id rest

        | [] -> k acc symtab
        | x :: _ -> raise (Bad_args (List.assoc x Stree.op_names))

let eval program =
        let finalise acc _ = fn_to_normal acc in
        execute_i finalise id Symtab.empty program

